/*
 * Copyright (c) 1998 - 2000 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * $Id: pop3-proxy.c,v 1.8 2000/07/03 16:23:53 lha Exp $
 */


#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <errno.h>
#include <fcntl.h>
#include <netinet/in.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <unistd.h>

#include <roken.h>
#include <err.h>

#include <des.h>
#include <krb.h>

#define KPOPPORT 1109
#define POP3PORT 110

enum { STATE_LOGIN, STATE_PASS, STATE_RUSER, STATE_RPASS, 
       STATE_PASSTHRU, STATE_DISCONN } state;


static int   kpop_port = 0;
static int   proxy_port = 0;
static char *kpop_host = NULL;
static char realm_name[REALM_SZ];

static int proxy_fd = -1;
static char *pop_name = "pop";

static char tf[MAXPATHLEN];

#ifndef STDIN_FD
#define STDIN_FD 0
#endif
#ifndef STDOUT_FD
#define STDOUT_FD 1
#endif



/*
 * fetch a line from `fd' and store it in `buf' that has size `sz'
 */

int 
getline(int fd, unsigned char *buf,int sz)
{
    int     x = 0;
    
    while(1) {
	if(read(fd,(char *)&buf[x],1) != 1)
	    return(-1);
	
	/* get \r\n - read a char and throw it away */
	if(buf[x] == '\r') {
	    if(read(fd,(char *)&buf[x],1) != 1)
		return(-1);
	    buf[x] = '\0';
	    return(x);
	}
	
	if(buf[x] == '\n') {
	    buf[x] = '\0';
	    return(x);
	}
	
	if(++x >= sz) {
	    return(-1);
	}
    }
}

/*
 * varargs verson of write
 */

void
writef (int fd, char *fmt, ...) 
{
    va_list args;
    char *str;
    int sz;

    va_start(args, fmt);
    
    sz = vasprintf (&str, fmt, args);

    va_end(args);

    if (sz == -1) {
	syslog (LOG_NOTICE, "writef: %d", sz) ;
	exit (0);
    }
	
    if (sz > 0) {
	if (write (fd, str, sz) != sz) {
	    syslog (LOG_NOTICE, "write: %d", sz) ;
	    exit (0);
	}
    }
    free (str);
}

/*
 * connect to the real pop3 server `host' on `port' with `user' and
 * password `passwd'.
 */

static int
conn_realserver (char *host, int port, char *user, char *passwd) 
{
    struct hostent *hostent;
    struct sockaddr_in addr;
    KTEXT ticket;
    char *realm;
    MSG_DAT msg_data;
    CREDENTIALS cred;
    Key_schedule schedule;
    int ret;
        
    assert (user && passwd);

    realm = strdup(krb_realmofhost (host));
    if (realm == NULL) {
	writef (STDOUT_FD, "-ERR can't find/malloc realm of host %s\n", host) ;
	return -1;
    }
    
    do{
	snprintf(tf, sizeof(tf),
		 "%s_pop_%u_%u",
		 TKT_ROOT,
		 (unsigned)getuid(),
		 (unsigned)(getpid()*time(0)));
	ret = open(tf, O_CREAT|O_EXCL|O_RDWR);
    }while(ret < 0);
    close(ret);
    unlink(tf);
    krb_set_tkt_string (tf);

    ret = krb_get_pw_in_tkt(user,
			    "",
			    realm_name,
			    KRB_TICKET_GRANTING_TICKET,
			    realm,
			    DEFAULT_TKT_LIFE,
			    passwd);
    
    memset (passwd, 0, strlen(passwd));

    if (ret != KSUCCESS) {
	writef (STDOUT_FD, "-ERR Incorrect password\n") ;
	return -1;
    }
    
    hostent = gethostbyname (host);
    
    if (hostent == NULL) {
	writef (STDOUT_FD, "-ERR can't resolve hostname %s\n", host) ;
	return -1;
    }

    memset (&addr, 0, sizeof (addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(port) ;
    
    proxy_fd = socket (PF_INET, SOCK_STREAM, 0);
    
    bcopy (*hostent->h_addr_list, &addr.sin_addr, sizeof(addr.sin_addr));

    ret = connect (proxy_fd, (struct sockaddr *) &addr, sizeof (addr));
    if (ret) {
	writef (STDOUT_FD, "-ERR can't connect to %s\n", host) ;
	return -1;
    }
    
    ticket = (KTEXT) malloc (sizeof (KTEXT_ST));
    if (ticket == NULL) {
	writef (STDOUT_FD, "-ERR conn_realserver: malloc\n") ;
	return -1;
    }

    
    ret = krb_sendauth (0L, proxy_fd, ticket, pop_name, host,
			realm,
			(unsigned long) 0, &msg_data, &cred, schedule,
			(struct sockaddr_in *) 0,
			(struct sockaddr_in *) 0,
			"KPOPV0.1");
    free (ticket);
    
    if (ret != KSUCCESS) {
	writef (STDOUT_FD, "-ERR conn_realserver: krb_sendauth: %d\n", ret) ;
	close (proxy_fd);
	return -1;
    }

    return 0;
}


static void
copyloop(int sock1, int sock2, int timeout_secs)
{
    fd_set ifds;
    int bytes ;
    int max_fd;                 /* Maximum numbered fd used */
    struct timeval timeout;
    char buf[BUFSIZ];
    int error;
    
    /* file descriptor bits */
    if (sock1 > sock2) {
	max_fd = sock1;
    } else {
	max_fd = sock2;
    }
    
    while(1) {
	FD_ZERO(&ifds);
	FD_SET(sock1, &ifds);
	FD_SET(sock2, &ifds);
	
	/* Set up timeout */
	timeout.tv_sec = timeout_secs;
	timeout.tv_usec = 0;
    
	error = select(max_fd + 1, &ifds, NULL, NULL, 
		   (timeout_secs ? &timeout : NULL));
	if (error < 0) {
	    if (errno == EINTR)
		continue;
	    break;
	}
	
	if(FD_ISSET(sock1, &ifds)) {
	    if((bytes = read (sock1, buf, sizeof(buf))) <= 0)
		break;
	    if(write (sock2, buf, bytes) != bytes)
		break;
	}
	if(FD_ISSET(sock2, &ifds)) {
	    if((bytes = read (sock2, buf, sizeof(buf))) <= 0)
		break;
	    if(write (sock1, buf, bytes) != bytes)
		break;
	}
    }
}

/*
 * Check if host is to be considered local
 */

static int
check_host(void) 
{
    struct sockaddr_in local ;
    struct sockaddr_in remote ;
    int rsize, lsize ;
    int ret;
    
    lsize = rsize = sizeof(local) ;

    ret = getpeername (STDIN_FD, (struct sockaddr *) &remote, &rsize);
    if (ret) {
	writef (STDIN_FD, "-ERR Error resolving remote addr: %d\n", errno) ;
	return -1;
    }

    ret = getsockname (STDIN_FD, (struct sockaddr *) &local, &lsize);
    if (ret) {
	writef (STDIN_FD, "-ERR Error resolving local addr %d\n", errno) ;
	return -1;
    }
    
    if (local.sin_addr.s_addr != remote.sin_addr.s_addr) {
	writef (STDIN_FD, "-ERR You can only connect from the local host\n") ;
	return 1;
    }
    
    return 0 ;
}


/*
 * user info
 */

static char *pop_user = NULL;
static char *pop_passwd = NULL;

/*
 * quit
 */

static int
quit_cmd (char *str, size_t sz)
{
    writef (STDOUT_FD, "+OK See you later dude\n");
    state = STATE_DISCONN;
    return 0;
}

/*
 * set username
 */

static int
user_cmd (char *str, size_t sz)
{
    if (sz < 2)
	return -1;

    if (state != STATE_LOGIN && state != STATE_PASS) {
	writef (STDOUT_FD, "-ERR can't give that command here\n");
	return 0;
    }
	
    str++;
    sz--;

    if (pop_user)
	free (pop_user);

    pop_user = strdup (str);
    if (pop_user) {
	writef (STDOUT_FD, "+OK Ok give password\n");
	state = STATE_PASS;
    } else
	writef (STDOUT_FD, "-ERR out of memory\n");

    return 0;
}

/*
 * set password
 */

static int
pass_cmd (char *str, size_t sz)
{
    if (sz < 2)
	return -1;

    if (state != STATE_PASS) {
	writef (STDOUT_FD, "-ERR can't give that command here\n");
	return 0;
    }
    str++;
    sz--;

    if (pop_passwd)
	free (pop_passwd);

    pop_passwd = strdup (str);
    if (pop_passwd) {

	if (conn_realserver (kpop_host, kpop_port, pop_user, pop_passwd))
	    return 0; /* something went wrong */

	state = STATE_RUSER;

	free (pop_passwd);
	pop_passwd = NULL;

    } else
	writef (STDOUT_FD, "-ERR out of memory\n");

    return 0;
}

/*
 * got a -ERR from remove server
 */

static int
remote_err_cmd (char *str, size_t sz)
{
    switch (state) {
    case STATE_RUSER:
    case STATE_RPASS:
	writef (STDOUT_FD, "-ERR%s", str);
	state = STATE_PASS;
	break;
    default:
	return -1;
    }
    return 0;
}

/*
 * got a +OK from remove server
 */

static int
remote_ok_cmd (char *str, size_t sz)
{
    switch (state) {
    case STATE_RUSER:
	assert (pop_user);
	writef (proxy_fd,  "USER %s\r\n", pop_user);
	free (pop_user);
	pop_user = NULL;
	state = STATE_RPASS;
	break;
    case STATE_RPASS:
	writef (proxy_fd,  "PASS hej\r\n");
	state = STATE_PASSTHRU;
	break;
    default:
	return -1;
    }
    return 0;
}


/*
 * comand table
 */

struct commands {
    char *cmd;
    int len;
    int (*proc) (char *str, size_t sz);
} cmds[] = {
    { "quit", 4, quit_cmd },
    { "user", 4, user_cmd },
    { "pass", 4, pass_cmd },
    { "-ERR", 4, remote_err_cmd },
    { "+OK", 3,  remote_ok_cmd },
    { NULL }
};

/*
 * parse input `cmdstr' and change state
 */

static int
parse_input(char *cmdstr, size_t sz) 
{
    struct commands *cmd = cmds;

    while (cmd->cmd) {
	if (strncasecmp (cmdstr, cmd->cmd, cmd->len) == 0)
	    return cmd->proc (cmdstr + cmd->len, sz - cmd->len);
	cmd++;
    }

    writef (STDOUT_FD, "-ERR no such command\n");

    return 0;
}


static void
doclientstuff(void) 
{
    int ret;
    char buf[BUFSIZ];

    ret = krb_get_lrealm (realm_name, 1);
    if (ret != KSUCCESS) {
	writef (STDOUT_FD, "-ERR can't find realm of this host\n") ;
	return;
    }

    writef (STDOUT_FD, 
	    "+OK POP-PROXY realm: %s - kpop: %s - kport: %d\n",
	    realm_name, kpop_host, kpop_port) ;    
    
    do {
	if (state == STATE_LOGIN || state == STATE_PASS)
	    ret = getline (STDIN_FD, buf, sizeof(buf));
	else
	    ret = getline (proxy_fd, buf, sizeof(buf));
	if (ret < 0)
	    break;

	if (parse_input (buf, ret))
	    writef (STDOUT_FD, "-ERR parse error\n") ;
	
    } while (state != STATE_DISCONN &&
	     state != STATE_PASSTHRU);

    if (state == STATE_PASSTHRU)
	copyloop (STDIN_FD, proxy_fd, 20);
}

int
main(int argc,char **argv)
{
    int do_inetdp = 0;
    int networkp = 1;

    char c;
    

    openlog ("pop3-proxy", LOG_PID, LOG_DAEMON);

    while ((c = getopt(argc, argv, "h:p:q:il")) != -1) {
	switch(c) {
	case 'h':
	    kpop_host = optarg;
	    break;
	case 'p':
	    kpop_port = atoi(optarg);
	    break;
	case 'q':
	    proxy_port = atoi(optarg);
	    break;
	case 'i':
	    do_inetdp = 1;
	    break;
	case 'l':
	    networkp = 0;
	    break;
	default:
	    errx(1, "unknown option");
	}
    }

    if (kpop_port == 0)
	kpop_port = KPOPPORT;

    if (proxy_port == 0)
	proxy_port = POP3PORT;

    if (kpop_host == NULL)
	errx (1, "-ERR you have to give a host to pop3-proxy");

    if (networkp && do_inetdp)
	mini_inetd(htons(proxy_port));

    if (networkp && check_host()) 
	return 0;

    doclientstuff() ;

    dest_tkt();

    return 0;
}
